@@ -10,7 +10,7 @@
  *
  * You should have received a copy of the GNU Lesser General Public License
  * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
  */
 
 #include "lib.h"
@@ -28,26 +28,41 @@
 static struct utsname _utsname;
 static int _utsinit = 0;
 
-static char *_format_pvsegs(struct dm_pool *mem, const struct lv_segment *seg,
-			    int range_format, int metadata_areas_only)
+int lv_is_historical(const struct logical_volume *lv)
+{
+	return lv->this_glv && lv->this_glv->is_historical;
+}
+
+static struct dm_list *_format_pvsegs(struct dm_pool *mem, const struct lv_segment *seg,
+				      int range_format, int metadata_areas_only,
+				      int mark_hidden)
 {
 	unsigned int s;
 	const char *name = NULL;
 	uint32_t extent = 0;
+	uint32_t seg_len = 0;
 	char extent_str[32];
+	struct logical_volume *lv;
+	int visible = 1;
+	char *list_item;
+	size_t list_item_len;
+	struct dm_list *result = NULL;
 
-	if (!dm_pool_begin_object(mem, 256)) {
-		log_error("dm_pool_begin_object failed");
-		return NULL;
+	if (!(result = str_list_create(mem))) {
+		log_error("_format_pvsegs: str_list_create failed");
+		goto bad;
 	}
 
-	if (metadata_areas_only && (!seg_is_raid(seg) || lv_is_raid_metadata(seg->lv) || lv_is_raid_image(seg->lv)))
+	if (metadata_areas_only && (!seg_is_raid_with_meta(seg) || !seg->meta_areas || lv_is_raid_metadata(seg->lv) || lv_is_raid_image(seg->lv)))
 		goto out;
 
 	for (s = 0; s < seg->area_count; s++) {
 		switch (metadata_areas_only ? seg_metatype(seg, s) : seg_type(seg, s)) {
 		case AREA_LV:
-			name = metadata_areas_only ? seg_metalv(seg, s)->name : seg_lv(seg, s)->name;
+			lv = metadata_areas_only ? seg_metalv(seg, s) : seg_lv(seg, s);
+			seg_len = metadata_areas_only ? seg_metalv(seg, s)->le_count : seg_lv(seg, s)->le_count;
+			visible = lv_is_visible(lv);
+			name = lv->name;
 			extent = metadata_areas_only ? seg_le(seg, s) : 0;
 			break;
 		case AREA_PV:
@@ -56,79 +71,141 @@ static char *_format_pvsegs(struct dm_pool *mem, const struct lv_segment *seg,
 				continue;
 			name = dev_name(seg_dev(seg, s));
 			extent = seg_pe(seg, s);
+			seg_len = seg->area_len;
 			break;
 		case AREA_UNASSIGNED:
 			name = "unassigned";
 			extent = 0;
+			seg_len = 0;
 			break;
 		default:
 			log_error(INTERNAL_ERROR "Unknown area segtype.");
-			return NULL;
+			goto bad;
 		}
 
-		if (!dm_pool_grow_object(mem, name, strlen(name))) {
-			log_error("dm_pool_grow_object failed");
-			return NULL;
-		}
-
-		if (dm_snprintf(extent_str, sizeof(extent_str),
-				"%s%" PRIu32 "%s",
-				range_format ? ":" : "(", extent,
-				range_format ? "-"  : ")") < 0) {
-			log_error("Extent number dm_snprintf failed");
-			return NULL;
-		}
-		if (!dm_pool_grow_object(mem, extent_str, strlen(extent_str))) {
-			log_error("dm_pool_grow_object failed");
-			return NULL;
-		}
+		list_item_len = strlen(name);
+		if (!visible && mark_hidden)
+			/* +2 for [ ] */
+			list_item_len += 2;
 
 		if (range_format) {
 			if (dm_snprintf(extent_str, sizeof(extent_str),
-					FMTu32, metadata_areas_only ? extent + seg_metalv(seg, s)->le_count - 1 : extent + seg->area_len - 1) < 0) {
-				log_error("Extent number dm_snprintf failed");
-				return NULL;
+					":%" PRIu32 "-%" PRIu32,
+					extent, extent + seg_len - 1) < 0) {
+				log_error("_format_pvseggs: extent range dm_snprintf failed");
+				goto bad;
 			}
-			if (!dm_pool_grow_object(mem, extent_str, strlen(extent_str))) {
-				log_error("dm_pool_grow_object failed");
-				return NULL;
+		} else {
+			if (dm_snprintf(extent_str, sizeof(extent_str),
+					"(%" PRIu32 ")", extent) < 0) {
+				log_error("_format_pvsegs: extent number dm_snprintf failed");
+				goto bad;
 			}
 		}
+		list_item_len += strlen(extent_str);
+		/* trialing 0 */
+		list_item_len += 1;
 
-		if ((s != seg->area_count - 1) &&
-		    !dm_pool_grow_object(mem, range_format ? " " : ",", 1)) {
-			log_error("dm_pool_grow_object failed");
-			return NULL;
+		if (!(list_item = dm_pool_zalloc(mem, list_item_len))) {
+			log_error("_format_pvsegs: list item dm_pool_zalloc failed");
+			goto bad;
+		}
+
+		if (dm_snprintf(list_item, list_item_len,
+				"%s%s%s%s",
+				(!visible && mark_hidden) ? "[" : "",
+				name,
+				(!visible && mark_hidden) ? "]" : "",
+				extent_str) < 0) {
+			log_error("_format_pvsegs: list item dmsnprintf failed");
+			goto bad;
+		}
+
+		if (!str_list_add_no_dup_check(mem, result, list_item)) {
+			log_error("_format_pvsegs: failed to add item to list");
+			goto bad;
 		}
 	}
-
 out:
-	if (!dm_pool_grow_object(mem, "\0", 1)) {
-		log_error("dm_pool_grow_object failed");
-		return NULL;
-	}
-
-	return dm_pool_end_object(mem);
+	return result;
+bad:
+	dm_pool_free(mem, result);
+	return NULL;
 }
 
-char *lvseg_devices(struct dm_pool *mem, const struct lv_segment *seg)
+struct dm_list *lvseg_devices(struct dm_pool *mem, const struct lv_segment *seg)
 {
-	return _format_pvsegs(mem, seg, 0, 0);
+	return _format_pvsegs(mem, seg, 0, 0, 0);
 }
 
-char *lvseg_metadata_devices(struct dm_pool *mem, const struct lv_segment *seg)
+char *lvseg_devices_str(struct dm_pool *mem, const struct lv_segment *seg)
 {
-	return _format_pvsegs(mem, seg, 0, 1);
+	struct dm_list *list;
+
+	if (!(list = lvseg_devices(mem, seg)))
+		return_NULL;
+
+	return str_list_to_str(mem, list, ",");
 }
 
-char *lvseg_seg_pe_ranges(struct dm_pool *mem, const struct lv_segment *seg)
+struct dm_list *lvseg_metadata_devices(struct dm_pool *mem, const struct lv_segment *seg)
 {
-	return _format_pvsegs(mem, seg, 1, 0);
+	return _format_pvsegs(mem, seg, 0, 1, 0);
 }
 
-char *lvseg_seg_metadata_le_ranges(struct dm_pool *mem, const struct lv_segment *seg)
+char *lvseg_metadata_devices_str(struct dm_pool *mem, const struct lv_segment *seg)
 {
-	return _format_pvsegs(mem, seg, 1, 1);
+	struct dm_list *list;
+
+	if (!(list = lvseg_devices(mem, seg)))
+		return_NULL;
+
+	return str_list_to_str(mem, list, ",");
+}
+
+struct dm_list *lvseg_seg_pe_ranges(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	return _format_pvsegs(mem, seg, 1, 0, 0);
+}
+
+char *lvseg_seg_pe_ranges_str(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	struct dm_list *list;
+
+	if (!(list = lvseg_seg_pe_ranges(mem, seg)))
+		return_NULL;
+
+	return str_list_to_str(mem, list, " ");
+}
+
+struct dm_list *lvseg_seg_le_ranges(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	return _format_pvsegs(mem, seg, 1, 0, seg->lv->vg->cmd->report_mark_hidden_devices);
+}
+
+char *lvseg_seg_le_ranges_str(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	struct dm_list *list;
+
+	if (!(list = lvseg_seg_pe_ranges(mem, seg)))
+		return_NULL;
+
+	return str_list_to_str(mem, list, seg->lv->vg->cmd->report_list_item_separator);
+}
+
+struct dm_list *lvseg_seg_metadata_le_ranges(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	return _format_pvsegs(mem, seg, 1, 1, seg->lv->vg->cmd->report_mark_hidden_devices);
+}
+
+char *lvseg_seg_metadata_le_ranges_str(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	struct dm_list *list;
+
+	if (!(list = lvseg_seg_metadata_le_ranges(mem, seg)))
+		return_NULL;
+
+	return str_list_to_str(mem, list, seg->lv->vg->cmd->report_list_item_separator);
 }
 
 char *lvseg_tags_dup(const struct lv_segment *seg)
@@ -146,6 +223,47 @@ char *lvseg_discards_dup(struct dm_pool *mem, const struct lv_segment *seg)
 	return  dm_pool_strdup(mem, get_pool_discards_name(seg->discards));
 }
 
+char *lvseg_kernel_discards_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_with_info_and_seg_status *lvdm)
+{
+	const char *s = "";
+	char *ret;
+
+	if (lvdm->seg_status.type == SEG_STATUS_THIN_POOL)
+		s = get_pool_discards_name(lvdm->seg_status.thin_pool->discards);
+
+	if (!(ret = dm_pool_strdup(mem, s))) {
+		log_error("lvseg_kernel_discards_dup_with_info_and_seg_status: dm_pool_strdup failed");
+		return NULL;
+	}
+
+	return ret;
+}
+
+char *lvseg_kernel_discards_dup(struct dm_pool *mem, const struct lv_segment *seg)
+{
+	char *ret = NULL;
+	struct lv_with_info_and_seg_status status = {
+		.seg_status = {
+			.type = SEG_STATUS_NONE,
+			.seg = seg
+		},
+	};
+
+	if (!lv_is_thin_pool(seg->lv))
+		return NULL;
+
+	if (!(status.seg_status.mem = dm_pool_create("reporter_pool", 1024)))
+		return_NULL;
+
+	if (!(status.info_ok = lv_info_with_seg_status(seg->lv->vg->cmd, seg->lv, seg, 1, &status, 0, 0)))
+		goto_bad;
+
+	ret = lvseg_kernel_discards_dup_with_info_and_seg_status(mem, &status);
+bad:
+	dm_pool_destroy(status.seg_status.mem);
+	return ret;
+}
+
 char *lvseg_cachemode_dup(struct dm_pool *mem, const struct lv_segment *seg)
 {
 	const char *name = get_cache_mode_name(seg);
@@ -227,6 +345,90 @@ uint64_t lvseg_size(const struct lv_segment *seg)
 	return (uint64_t) seg->len * seg->lv->vg->extent_size;
 }
 
+dm_percent_t lvseg_percent_with_info_and_seg_status(const struct lv_with_info_and_seg_status *lvdm,
+						    percent_get_t type)
+{
+	dm_percent_t p;
+	uint64_t csize;
+	const struct lv_segment *seg;
+	const struct lv_seg_status *s = &lvdm->seg_status;
+
+	/*
+	 * TODO:
+	 *   Later move to segment methods, instead of using single place.
+	 *   Also handle logic for mirror segments and it total_* summing
+	 *   Esentially rework  _target_percent API for segtype.
+	 */
+	switch (s->type) {
+	case SEG_STATUS_CACHE:
+		if (s->cache->fail || s->cache->error)
+			p = DM_PERCENT_INVALID;
+		else {
+			switch (type) {
+			case PERCENT_GET_DIRTY:
+				p = dm_make_percent(s->cache->dirty_blocks,
+						    s->cache->used_blocks);
+				break;
+			case PERCENT_GET_METADATA:
+				p = dm_make_percent(s->cache->metadata_used_blocks,
+						    s->cache->metadata_total_blocks);
+				break;
+			default:
+				p = dm_make_percent(s->cache->used_blocks,
+						    s->cache->total_blocks);
+			}
+		}
+		break;
+	case SEG_STATUS_SNAPSHOT:
+		if (s->snapshot->merge_failed)
+			p = DM_PERCENT_INVALID;
+		else if (s->snapshot->invalid)
+			p = DM_PERCENT_100; /* Shown as 100% full */
+		else if (s->snapshot->has_metadata_sectors &&
+			 (s->snapshot->used_sectors == s->snapshot->metadata_sectors))
+			p = DM_PERCENT_0;
+		else
+			p = dm_make_percent(s->snapshot->used_sectors,
+					    s->snapshot->total_sectors);
+		break;
+	case SEG_STATUS_THIN_POOL:
+		if (s->thin_pool->fail || s->thin_pool->error)
+			p = DM_PERCENT_INVALID;
+		else if (type == PERCENT_GET_METADATA)
+			p = dm_make_percent(s->thin_pool->used_metadata_blocks,
+					    s->thin_pool->total_metadata_blocks);
+		else
+			p = dm_make_percent(s->thin_pool->used_data_blocks,
+					    s->thin_pool->total_data_blocks);
+		break;
+	case SEG_STATUS_THIN:
+		if (s->thin->fail || (type != PERCENT_GET_DATA))
+			/* TODO: expose highest mapped sector */
+			p = DM_PERCENT_INVALID;
+		else {
+			seg = first_seg(lvdm->lv);
+			/* Pool allocates whole chunk so round-up to nearest one */
+			csize = first_seg(seg->pool_lv)->chunk_size;
+			csize = ((seg->lv->size + csize - 1) / csize) * csize;
+			if (s->thin->mapped_sectors <= csize)
+				p = dm_make_percent(s->thin->mapped_sectors, csize);
+			else {
+				log_warn("WARNING: Thin volume %s maps %s while the size is only %s.",
+					 display_lvname(seg->lv),
+					 display_size(lvdm->lv->vg->cmd, s->thin->mapped_sectors),
+					 display_size(lvdm->lv->vg->cmd, csize));
+				/* Don't show nonsense numbers like i.e. 1000% full */
+				p = DM_PERCENT_100;
+			}
+		}
+		break;
+	default:
+		p = DM_PERCENT_INVALID;
+	}
+
+	return p;
+}
+
 uint32_t lv_kernel_read_ahead(const struct logical_volume *lv)
 {
 	struct lvinfo info;
@@ -236,11 +438,101 @@ uint32_t lv_kernel_read_ahead(const struct logical_volume *lv)
 	return info.read_ahead;
 }
 
-static char *_do_lv_origin_dup(struct dm_pool *mem, const struct logical_volume *lv,
-			       int uuid)
-{
-	struct logical_volume *origin;
+struct pv_and_int {
+	struct physical_volume *pv;
+	int *i;
+};
 
+static int _lv_is_on_pv(struct logical_volume *lv, void *data)
+{
+	int *is_on_pv = ((struct pv_and_int *)data)->i;
+	struct physical_volume *pv = ((struct pv_and_int *)data)->pv;
+	uint32_t s;
+	struct physical_volume *pv2;
+	struct lv_segment *seg;
+
+	if (!lv || !(first_seg(lv)))
+		return_0;
+
+	/*
+	 * If the LV has already been found to be on the PV, then
+	 * we don't need to continue checking - just return.
+	 */
+	if (*is_on_pv)
+		return 1;
+
+	dm_list_iterate_items(seg, &lv->segments) {
+		for (s = 0; s < seg->area_count; s++) {
+			if (seg_type(seg, s) != AREA_PV)
+				continue;
+
+			pv2 = seg_pv(seg, s);
+			if (id_equal(&pv->id, &pv2->id)) {
+				*is_on_pv = 1;
+				return 1;
+			}
+			if (pv->dev && pv2->dev &&
+			    (pv->dev->dev == pv2->dev->dev)) {
+				*is_on_pv = 1;
+				return 1;
+			}
+		}
+	}
+
+	return 1;
+}
+
+/*
+ * lv_is_on_pv
+ * @lv:
+ * @pv:
+ *
+ * If any of the component devices of the LV are on the given PV, 1
+ * is returned; otherwise 0.  For example if one of the images of a RAID
+ * (or its metadata device) is on the PV, 1 would be returned for the
+ * top-level LV.
+ * If you wish to check the images themselves, you should pass them.
+ *
+ * Returns: 1 if LV (or part of LV) is on PV, 0 otherwise
+ */
+int lv_is_on_pv(struct logical_volume *lv, struct physical_volume *pv)
+{
+	int is_on_pv = 0;
+	struct pv_and_int context = { pv, &is_on_pv };
+
+	if (!_lv_is_on_pv(lv, &context) ||
+	    !for_each_sub_lv(lv, _lv_is_on_pv, &context))
+		/* Failure only happens if bad arguments are passed */
+		log_error(INTERNAL_ERROR "for_each_sub_lv failure.");
+
+	log_debug_metadata("%s is %son %s", lv->name,
+			   is_on_pv ? "" : "not ", pv_dev_name(pv));
+	return is_on_pv;
+}
+
+/*
+ * lv_is_on_pvs
+ * @lv
+ * @pvs
+ *
+ * Returns 1 if the LV (or part of the LV) is on any of the pvs
+ * in the list, 0 otherwise.
+ */
+int lv_is_on_pvs(struct logical_volume *lv, struct dm_list *pvs)
+{
+	struct pv_list *pvl;
+
+	dm_list_iterate_items(pvl, pvs)
+		if (lv_is_on_pv(lv, pvl->pv))
+			return 1;
+
+	return 0;
+}
+
+
+struct logical_volume *lv_origin_lv(const struct logical_volume *lv)
+{
+	struct logical_volume *origin = NULL;
 
 	if (lv_is_cow(lv))
 		origin = origin_from_cow(lv);
@@ -250,13 +542,22 @@ static char *_do_lv_origin_dup(struct dm_pool *mem, const struct logical_volume
 		origin = first_seg(lv)->origin;
 	else if (lv_is_thin_volume(lv) && first_seg(lv)->external_lv)
 		origin = first_seg(lv)->external_lv;
-	else
+
+	return origin;
+}
+
+static char *_do_lv_origin_dup(struct dm_pool *mem, const struct logical_volume *lv,
+			       int uuid)
+{
+	struct logical_volume *origin_lv = lv_origin_lv(lv);
+
+	if (!origin_lv)
 		return NULL;
 
 	if (uuid)
-		return lv_uuid_dup(mem, origin);
+		return lv_uuid_dup(mem, origin_lv);
 	else
-		return lv_name_dup(mem, origin);
+		return lv_name_dup(mem, origin_lv);
 }
 
 char *lv_origin_dup(struct dm_pool *mem, const struct logical_volume *lv)
@@ -276,30 +577,32 @@ char *lv_name_dup(struct dm_pool *mem, const struct logical_volume *lv)
 
 char *lv_fullname_dup(struct dm_pool *mem, const struct logical_volume *lv)
 {
-        char lvfullname[NAME_LEN * 2 + 2];
+	char lvfullname[NAME_LEN * 2 + 2];
 
-        if (dm_snprintf(lvfullname, sizeof(lvfullname), "%s/%s", lv->vg->name, lv->name) < 0) {
-                log_error("lvfullname snprintf failed");
-                return NULL;
-        }
+	if (dm_snprintf(lvfullname, sizeof(lvfullname), "%s/%s", lv->vg->name, lv->name) < 0) {
+		log_error("lvfullname snprintf failed");
+		return NULL;
+	}
 
-        return dm_pool_strdup(mem, lvfullname);
+	return dm_pool_strdup(mem, lvfullname);
 }
 
 struct logical_volume *lv_parent(const struct logical_volume *lv)
 {
 	struct logical_volume *parent_lv = NULL;
+	struct lv_segment *seg;
 
 	if (lv_is_visible(lv))
 		;
-	else if (lv_is_mirror_image(lv) || lv_is_mirror_log(lv))
-		parent_lv = get_only_segment_using_this_lv(lv)->lv;
-	else if (lv_is_raid_image(lv) || lv_is_raid_metadata(lv))
-		parent_lv = get_only_segment_using_this_lv(lv)->lv;
-	else if (lv_is_cache_pool_data(lv) || lv_is_cache_pool_metadata(lv))
-		parent_lv = get_only_segment_using_this_lv(lv)->lv;
-	else if (lv_is_thin_pool_data(lv) || lv_is_thin_pool_metadata(lv))
-		parent_lv = get_only_segment_using_this_lv(lv)->lv;
+	else if ((lv_is_mirror_image(lv) || lv_is_mirror_log(lv)) ||
+		 (lv_is_raid_image(lv) || lv_is_raid_metadata(lv)) ||
+		 (lv_is_cache_pool_data(lv) || lv_is_cache_pool_metadata(lv)) ||
+		 (lv_is_thin_pool_data(lv) || lv_is_thin_pool_metadata(lv))) {
+		if (!(seg = get_only_segment_using_this_lv(lv)))
+			stack;
+		else
+			parent_lv = seg->lv;
+	}
 
 	return parent_lv;
 }
@@ -325,23 +628,32 @@ char *lv_modules_dup(struct dm_pool *mem, const struct logical_volume *lv)
 	return tags_format_and_copy(mem, modules);
 }
 
-static char *_do_lv_mirror_log_dup(struct dm_pool *mem, const struct logical_volume *lv,
-				   int uuid)
+struct logical_volume *lv_mirror_log_lv(const struct logical_volume *lv)
 {
 	struct lv_segment *seg;
 
 	dm_list_iterate_items(seg, &lv->segments) {
-		if (seg_is_mirrored(seg) && seg->log_lv) {
-			if (uuid)
-				return lv_uuid_dup(mem, seg->log_lv);
-			else
-				return lv_name_dup(mem, seg->log_lv);
-		}
+		if (seg_is_mirrored(seg) && seg->log_lv)
+			return seg->log_lv;
 	}
 
 	return NULL;
 }
 
+static char *_do_lv_mirror_log_dup(struct dm_pool *mem, const struct logical_volume *lv,
+				   int uuid)
+{
+	struct logical_volume *mirror_log_lv = lv_mirror_log_lv(lv);
+
+	if (!mirror_log_lv)
+		return NULL;
+
+	if (uuid)
+		return lv_uuid_dup(mem, mirror_log_lv);
+	else
+		return lv_name_dup(mem, mirror_log_lv);
+}
+
 char *lv_mirror_log_dup(struct dm_pool *mem, const struct logical_volume *lv)
 {
 	return _do_lv_mirror_log_dup(mem, lv, 0);
@@ -352,22 +664,27 @@ char *lv_mirror_log_uuid_dup(struct dm_pool *mem, const struct logical_volume *l
 	return _do_lv_mirror_log_dup(mem, lv, 1);
 }
 
+struct logical_volume *lv_pool_lv(const struct logical_volume *lv)
+{
+	struct lv_segment *seg = (lv_is_thin_volume(lv) || lv_is_cache(lv)) ?
+				  first_seg(lv) : NULL;
+	struct logical_volume *pool_lv = seg ? seg->pool_lv : NULL;
+
+	return pool_lv;
+}
+
 static char *_do_lv_pool_lv_dup(struct dm_pool *mem, const struct logical_volume *lv,
 				int uuid)
 {
-	struct lv_segment *seg;
+	struct logical_volume *pool_lv = lv_pool_lv(lv);
 
-	dm_list_iterate_items(seg, &lv->segments) {
-		if (seg->pool_lv &&
-		    (seg_is_thin_volume(seg) || seg_is_cache(seg))) {
-			if (uuid)
-				return lv_uuid_dup(mem, seg->pool_lv);
-			else
-				return lv_name_dup(mem, seg->pool_lv);
-		}
-	}
+	if (!pool_lv)
+		return NULL;
 
-	return NULL;
+	if (uuid)
+		return lv_uuid_dup(mem, pool_lv);
+	else
+		return lv_name_dup(mem, pool_lv);
 }
 
 char *lv_pool_lv_dup(struct dm_pool *mem, const struct logical_volume *lv)
@@ -380,20 +697,27 @@ char *lv_pool_lv_uuid_dup(struct dm_pool *mem, const struct logical_volume *lv)
 	return _do_lv_pool_lv_dup(mem, lv, 1);
 }
 
+struct logical_volume *lv_data_lv(const struct logical_volume *lv)
+{
+	struct lv_segment *seg = (lv_is_thin_pool(lv) || lv_is_cache_pool(lv)) ?
+				  first_seg(lv) : NULL;
+	struct logical_volume *data_lv = seg ? seg_lv(seg, 0) : NULL;
+
+	return data_lv;
+}
+
 static char *_do_lv_data_lv_dup(struct dm_pool *mem, const struct logical_volume *lv,
 				int uuid)
 {
-	struct lv_segment *seg = (lv_is_thin_pool(lv) || lv_is_cache_pool(lv)) ?
-		first_seg(lv) : NULL;
+	struct logical_volume *data_lv = lv_data_lv(lv);
 
-	if (seg) {
-		if (uuid)
-			return lv_uuid_dup(mem, seg_lv(seg, 0));
-		else
-			return lv_name_dup(mem, seg_lv(seg, 0));
-	}
+	if (!data_lv)
+		return NULL;
 
-	return NULL;
+	if (uuid)
+		return lv_uuid_dup(mem, data_lv);
+	else
+		return lv_name_dup(mem, data_lv);
 }
 
 char *lv_data_lv_dup(struct dm_pool *mem, const struct logical_volume *lv)
@@ -406,21 +730,27 @@ char *lv_data_lv_uuid_dup(struct dm_pool *mem, const struct logical_volume *lv)
 	return _do_lv_data_lv_dup(mem, lv, 1);
 }
 
+struct logical_volume *lv_metadata_lv(const struct logical_volume *lv)
+{
+	struct lv_segment *seg = (lv_is_thin_pool(lv) || lv_is_cache_pool(lv)) ?
+				  first_seg(lv) : NULL;
+	struct logical_volume *metadata_lv = seg ? seg->metadata_lv : NULL;
+
+	return metadata_lv;
+}
 
 static char *_do_lv_metadata_lv_dup(struct dm_pool *mem, const struct logical_volume *lv,
 				    int uuid)
 {
-	struct lv_segment *seg = (lv_is_thin_pool(lv) || lv_is_cache_pool(lv)) ?
-		first_seg(lv) : NULL;
+	struct logical_volume *metadata_lv = lv_metadata_lv(lv);
 
-	if (seg) {
-		if (uuid)
-			return lv_uuid_dup(mem, seg->metadata_lv);
-		else
-			return lv_name_dup(mem, seg->metadata_lv);
-	}
+	if (!metadata_lv)
+		return NULL;
 
-	return NULL;
+	if (uuid)
+		return lv_uuid_dup(mem, metadata_lv);
+	else
+		return lv_name_dup(mem, metadata_lv);
 }
 
 char *lv_metadata_lv_dup(struct dm_pool *mem, const struct logical_volume *lv)
@@ -460,8 +790,7 @@ int lv_kernel_major(const struct logical_volume *lv)
 	return -1;
 }
 
-static char *_do_lv_convert_lv_dup(struct dm_pool *mem, const struct logical_volume *lv,
-				   int uuid)
+struct logical_volume *lv_convert_lv(const struct logical_volume *lv)
 {
 	struct lv_segment *seg;
 
@@ -470,16 +799,27 @@ static char *_do_lv_convert_lv_dup(struct dm_pool *mem, const struct logical_vol
 
 		/* Temporary mirror is always area_num == 0 */
 		if (seg_type(seg, 0) == AREA_LV &&
-		    is_temporary_mirror_layer(seg_lv(seg, 0))) {
-			if (uuid)
-				return lv_uuid_dup(mem, seg_lv(seg, 0));
-			else
-				return lv_name_dup(mem, seg_lv(seg, 0));
-		}
+		    is_temporary_mirror_layer(seg_lv(seg, 0)))
+			return seg_lv(seg, 0);
 	}
+
 	return NULL;
 }
 
+static char *_do_lv_convert_lv_dup(struct dm_pool *mem, const struct logical_volume *lv,
+				   int uuid)
+{
+	struct logical_volume *convert_lv = lv_convert_lv(lv);
+
+	if (!convert_lv)
+		return NULL;
+
+	if (uuid)
+		return lv_uuid_dup(mem, convert_lv);
+	else
+		return lv_name_dup(mem, convert_lv);
+}
+
 char *lv_convert_lv_dup(struct dm_pool *mem, const struct logical_volume *lv)
 {
 	return _do_lv_convert_lv_dup(mem, lv, 0);
@@ -590,7 +930,7 @@ char *lv_dmpath_dup(struct dm_pool *mem, const struct logical_volume *lv)
 	if (!*lv->vg->name)
 		return dm_pool_strdup(mem, "");
 
-        if (!(name = dm_build_dm_name(mem, lv->vg->name, lv->name, NULL))) {
+	if (!(name = dm_build_dm_name(mem, lv->vg->name, lv->name, NULL))) {
 		log_error("dm_build_dm_name failed");
 		return NULL;
 	}
@@ -756,7 +1096,7 @@ int lv_raid_healthy(const struct logical_volume *lv)
 	/* Find out which sub-LV this is. */
 	for (s = 0; s < raid_seg->area_count; s++)
 		if ((lv_is_raid_image(lv) && (seg_lv(raid_seg, s) == lv)) ||
-		    (lv_is_raid_metadata(lv) && (seg_metalv(raid_seg,s) == lv)))
+		    (lv_is_raid_metadata(lv) && (seg_metalv(raid_seg, s) == lv)))
 			break;
 	if (s == raid_seg->area_count) {
 		log_error(INTERNAL_ERROR
@@ -773,7 +1113,6 @@ int lv_raid_healthy(const struct logical_volume *lv)
 
 char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_with_info_and_seg_status *lvdm)
 {
-	dm_percent_t snap_percent;
 	const struct logical_volume *lv = lvdm->lv;
 	struct lv_segment *seg;
 	char *repstr;
@@ -784,7 +1123,7 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 	}
 
 	/* Blank if this is a "free space" LV. */
-	if (!*lv->name)
+	if (!*lv->name && !lv_is_historical(lv))
 		goto out;
 
 	if (lv_is_pvmove(lv))
@@ -848,7 +1187,10 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 
 	repstr[3] = (lv->status & FIXED_MINOR) ? 'm' : '-';
 
-	if (!activation() || !lvdm->info_ok) {
+	if (lv_is_historical(lv)) {
+		repstr[4] = 'h';
+		repstr[5] = '-';
+	} else if (!activation() || !lvdm->info_ok) {
 		repstr[4] = 'X';		/* Unknown */
 		repstr[5] = 'X';		/* Unknown */
 	} else if (lvdm->info.exists) {
@@ -862,22 +1204,29 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 			repstr[4] = 'd';	/* Inactive without table */
 
 		/* Snapshot dropped? */
-		if (lvdm->info.live_table && lv_is_cow(lv)) {
-			if (!lv_snapshot_percent(lv, &snap_percent) ||
-			    snap_percent == DM_PERCENT_INVALID) {
+		if (lvdm->info.live_table &&
+		    (lvdm->seg_status.type == SEG_STATUS_SNAPSHOT)) {
+			if (lvdm->seg_status.snapshot->invalid) {
 				if (lvdm->info.suspended)
 					repstr[4] = 'S'; /* Susp Inv snapshot */
 				else
 					repstr[4] = 'I'; /* Invalid snapshot */
-			}
-			else if (snap_percent == LVM_PERCENT_MERGE_FAILED) {
+			} else if (lvdm->seg_status.snapshot->merge_failed) {
 				if (lvdm->info.suspended)
 					repstr[4] = 'M'; /* Susp snapshot merge failed */
 				else
-					repstr[4] = 'm'; /* snapshot merge failed */
+					repstr[4] = 'm'; /* Snapshot merge failed */
 			}
 		}
 
+		/* 'c' when cache/thin-pool is active with needs_check flag
+		 * 'C' for suspend */
+		if ((lv_is_thin_pool(lv) &&
+		     lvdm->seg_status.thin_pool->needs_check) ||
+		    (lv_is_cache(lv) &&
+		     lvdm->seg_status.cache->needs_check))
+			repstr[4] = lvdm->info.suspended ? 'C' : 'c';
+
 		/*
 		 * 'R' indicates read-only activation of a device that
 		 * does not have metadata flagging it as read-only.
@@ -916,7 +1265,7 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 		repstr[7] = '-';
 
 	repstr[8] = '-';
-	if (lv->status & PARTIAL_LV)
+	if (lv_is_partial(lv))
 		repstr[8] = 'p';
 	else if (lv_is_raid_type(lv)) {
 		uint64_t n;
@@ -929,6 +1278,14 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 				repstr[8] = 'm';  /* RAID has 'm'ismatches */
 		} else if (lv->status & LV_WRITEMOSTLY)
 			repstr[8] = 'w';  /* sub-LV has 'w'ritemostly */
+	} else if (lv_is_cache(lv) &&
+		   (lvdm->seg_status.type != SEG_STATUS_NONE)) {
+		if (lvdm->seg_status.type == SEG_STATUS_UNKNOWN)
+			repstr[8] = 'X'; /* Unknown */
+		else if (lvdm->seg_status.cache->fail)
+			repstr[8] = 'F';
+		else if (lvdm->seg_status.cache->read_only)
+			repstr[8] = 'M';
 	} else if (lv_is_thin_pool(lv) &&
 		   (lvdm->seg_status.type != SEG_STATUS_NONE)) {
 		if (lvdm->seg_status.type == SEG_STATUS_UNKNOWN)
@@ -939,6 +1296,12 @@ char *lv_attr_dup_with_info_and_seg_status(struct dm_pool *mem, const struct lv_
 			repstr[8] = 'D';
 		else if (lvdm->seg_status.thin_pool->read_only)
 			repstr[8] = 'M';
+	} else if (lv_is_thin_volume(lv) &&
+		   (lvdm->seg_status.type != SEG_STATUS_NONE)) {
+		if (lvdm->seg_status.type == SEG_STATUS_UNKNOWN)
+			repstr[8] = 'X'; /* Unknown */
+		else if (lvdm->seg_status.thin->fail)
+			repstr[8] = 'F';
 	}
 
 	if (lv->status & LV_ACTIVATION_SKIP)
@@ -1006,12 +1369,13 @@ int lv_set_creation(struct logical_volume *lv,
 	return 1;
 }
 
-char *lv_time_dup(struct dm_pool *mem, const struct logical_volume *lv, int iso_mode)
+static char *_time_dup(struct cmd_context *cmd, struct dm_pool *mem,
+		       time_t ts, int iso_mode)
 {
 	char buffer[4096];
 	struct tm *local_tm;
-	time_t ts = (time_t)lv->timestamp;
-	const char *format = iso_mode ? DEFAULT_TIME_FORMAT : lv->vg->cmd->time_format;
+	const char *format = iso_mode ? DEFAULT_TIME_FORMAT
+				      : cmd->time_format;
 
 	if (!ts ||
 	    !(local_tm = localtime(&ts)) ||
@@ -1021,6 +1385,22 @@ char *lv_time_dup(struct dm_pool *mem, const struct logical_volume *lv, int iso_
 	return dm_pool_strdup(mem, buffer);
 }
 
+char *lv_creation_time_dup(struct dm_pool *mem, const struct logical_volume *lv, int iso_mode)
+{
+	time_t ts = lv_is_historical(lv) ? (time_t) lv->this_glv->historical->timestamp
+					 : (time_t) lv->timestamp;
+
+	return _time_dup(lv->vg->cmd, mem, ts, iso_mode);
+}
+
+char *lv_removal_time_dup(struct dm_pool *mem, const struct logical_volume *lv, int iso_mode)
+{
+	time_t ts = lv_is_historical(lv) ? (time_t)lv->this_glv->historical->timestamp_removed
+					 : (time_t)0;
+
+	return _time_dup(lv->vg->cmd, mem, ts, iso_mode);
+}
+
 char *lv_host_dup(struct dm_pool *mem, const struct logical_volume *lv)
 {
 	return dm_pool_strdup(mem, lv->hostname ? : "");
@@ -1052,49 +1432,50 @@ int lv_active_change(struct cmd_context *cmd, struct logical_volume *lv,
 	
 	if (is_change_activating(activate) &&
 	    !lockd_lv(cmd, lv, ay_with_mode, LDLV_PERSISTENT)) {
-		log_error("Failed to lock logical volume %s/%s", lv->vg->name, lv->name);
+		log_error("Failed to lock logical volume %s.", display_lvname(lv));
 		return 0;
 	}
 
 	switch (activate) {
 	case CHANGE_AN:
 deactivate:
-		log_verbose("Deactivating logical volume \"%s\"", lv->name);
+		log_verbose("Deactivating logical volume %s.", display_lvname(lv));
 		if (!deactivate_lv(cmd, lv))
 			return_0;
 		break;
 	case CHANGE_ALN:
 		if (vg_is_clustered(lv->vg) && (needs_exclusive || _lv_is_exclusive(lv))) {
 			if (!lv_is_active_locally(lv)) {
-				log_error("Cannot deactivate remotely exclusive device locally.");
+				log_error("Cannot deactivate remotely exclusive device %s locally.",
+					  display_lvname(lv));
 				return 0;
 			}
 			/* Unlock whole exclusive activation */
 			goto deactivate;
 		}
-		log_verbose("Deactivating logical volume \"%s\" locally.",
-			    lv->name);
+		log_verbose("Deactivating logical volume %s locally.",
+			    display_lvname(lv));
 		if (!deactivate_lv_local(cmd, lv))
 			return_0;
 		break;
 	case CHANGE_ALY:
 	case CHANGE_AAY:
 		if (needs_exclusive || _lv_is_exclusive(lv)) {
-			log_verbose("Activating logical volume \"%s\" exclusively locally.",
-				    lv->name);
+			log_verbose("Activating logical volume %s exclusively locally.",
+				    display_lvname(lv));
 			if (!activate_lv_excl_local(cmd, lv))
 				return_0;
 		} else {
-			log_verbose("Activating logical volume \"%s\" locally.",
-				    lv->name);
+			log_verbose("Activating logical volume %s locally.",
+				    display_lvname(lv));
 			if (!activate_lv_local(cmd, lv))
 				return_0;
 		}
 		break;
 	case CHANGE_AEY:
 exclusive:
-		log_verbose("Activating logical volume \"%s\" exclusively.",
-			    lv->name);
+		log_verbose("Activating logical volume %s exclusively.",
+			    display_lvname(lv));
 		if (!activate_lv_excl(cmd, lv))
 			return_0;
 		break;
@@ -1103,14 +1484,14 @@ exclusive:
 	default:
 		if (needs_exclusive || _lv_is_exclusive(lv))
 			goto exclusive;
-		log_verbose("Activating logical volume \"%s\".", lv->name);
+		log_verbose("Activating logical volume %s.", display_lvname(lv));
 		if (!activate_lv(cmd, lv))
 			return_0;
 	}
 
 	if (!is_change_activating(activate) &&
 	    !lockd_lv(cmd, lv, "un", LDLV_PERSISTENT))
-		log_error("Failed to unlock logical volume %s/%s", lv->vg->name, lv->name);
+		log_error("Failed to unlock logical volume %s.", display_lvname(lv));
 
 	return 1;
 }
@@ -1169,7 +1550,8 @@ const struct logical_volume *lv_lock_holder(const struct logical_volume *lv)
 		/* Find any active LV from the pool */
 		dm_list_iterate_items(sl, &lv->segs_using_this_lv)
 			if (lv_is_active(sl->seg->lv)) {
-				log_debug("Thin volume \"%s\" is active.", sl->seg->lv->name);
+				log_debug_activation("Thin volume %s is active.",
+						     display_lvname(lv));
 				return sl->seg->lv;
 			}
 
@@ -1199,3 +1581,19 @@ struct profile *lv_config_profile(const struct logical_volume *lv)
 {
 	return lv->profile ? : lv->vg->profile;
 }
+
+int lv_has_constant_stripes(struct logical_volume *lv)
+{
+	uint32_t previous_area_count = 0;
+	struct lv_segment *seg;
+
+	dm_list_iterate_items(seg, &lv->segments) {
+		if (!seg_is_striped(seg))
+			return 0;
+		if (previous_area_count && previous_area_count != seg->area_count)
+			return 0;
+		previous_area_count = seg->area_count;
+	}
+
+	return 1;
+}
